import logging
from time import sleep
from typing import Any, Callable

from .waiter_src.comparators import Comparator, is_equal
from .waiter_src.exceptions import WaiterConditionWasNotMet

logger = logging.getLogger(__name__)


def retry(
    expected: Any, *, comparer: Comparator = is_equal, loops: int = 60, delay: int = 1
):
    """
    Simple decorator, that retries function if its result is different from expected.

    Example:
        ```python
            @bepatient.retry(200)
            def send_request():
                return requests.get('https://webludus.pl/en').status_code

            result = send_request()
            assert result == 200
        ```
    """

    def wrap(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
        def wrapped(*args, **kwargs) -> Any:
            for attempt in range(loops):
                logger.info(
                    "Checking whether the condition has been met. The %s approach."
                    " Expected: %s",
                    attempt + 1,
                    expected,
                )
                result = func(*args, **kwargs)
                if comparer(expected, result):
                    logger.info("Condition met! Result: %s", result)
                    return result
                logger.info(
                    "Condition was not met! Expected: %s | Result %s", expected, result
                )
                sleep(delay)
            raise WaiterConditionWasNotMet()

        return wrapped

    return wrap



import logging
from abc import ABC, abstractmethod
from typing import Any, Callable

log = logging.getLogger(__name__)


class Checker(ABC):
    """An abstract class defining the interface for a checker to be used by a Waiter."""

    def __init__(self, comparer: Callable[[Any, Any], bool], expected_value: Any):
        self.comparer = comparer
        self.expected_value = expected_value
        self._prepared_data: Any = None

    def __str__(self) -> str:
        """Textual representation of the Checker object for logging"""
        attrs = self.__dict__.copy()
        del attrs["_prepared_data"]
        attrs["checker"] = self.__class__.__name__
        attrs["comparer"] = self.comparer.__name__

        return (
            " | ".join([f"{k.capitalize()}: {v}" for k, v in sorted(attrs.items())])
            + f" | Data: {self._prepared_data}"
        )

    @abstractmethod
    def prepare_data(self, data: Any, run_uuid: str | None = None) -> Any:
        """Prepare the data from the response for comparison.

        Args:
            data (Response): response containing the data.
            run_uuid (str | None): unique run identifier. Defaults to None.

        Returns:
            Any: Data for comparison."""

    def check(self, data: Any, run_uuid: str) -> bool:
        """Check if the given data meets a certain condition.

        Args:
            data (Any): The data to be checked.
            run_uuid (str): unique run identifier.

        Returns:
            bool: True if the condition is met, False otherwise."""
        log.debug("Check uuid: %s | %s", run_uuid, self)

        self._prepared_data = self.prepare_data(data, run_uuid)
        if self.comparer(self._prepared_data, self.expected_value):
            log.info(
                "Check success! | uuid: %s | %s",
                run_uuid,
                self,
            )
            return True
        log.info("Check uuid: %s | Condition not met | %s", run_uuid, self)
        return False



from collections.abc import Iterable, Sized
from typing import Any, Callable, Literal, TypeAlias


def is_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is equal to expected_value, False otherwise."""
    return data == expected_value


def is_not_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is not equal to expected_value, False otherwise."""
    return data != expected_value


def is_greater_than(data: Any, expected_value: Any) -> bool:
    """Returns True if data is greater than expected_value, False otherwise."""
    try:
        return data > expected_value
    except TypeError:
        return False


def is_lesser_than(data: Any, expected_value: Any) -> bool:
    """Returns True if data is lesser than expected_value, False otherwise."""
    try:
        return data < expected_value
    except TypeError:
        return False


def is_greater_than_or_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is greater than or equal to expected_value,
    False otherwise."""
    try:
        return data >= expected_value
    except TypeError:
        return False


def is_lesser_than_or_equal(data: Any, expected_value: Any) -> bool:
    """Returns True if data is lesser than or equal to expected_value,
    False otherwise."""
    try:
        return data <= expected_value
    except TypeError:
        return False


def contain(data: Iterable[Any], expected_value: Any) -> bool:
    """Returns True if expected_value is present in data, False otherwise."""
    try:
        return expected_value in data
    except TypeError:
        return False


def not_contain(data: Iterable[Any], expected_value: Any) -> bool:
    """Returns True if expected_value is not present in data, False otherwise."""
    try:
        return expected_value not in data
    except TypeError:
        return False


def contain_all(data: Iterable[Any], expected_value: Iterable[Any]) -> bool:
    """Returns True if all elements in expected_value are present in data,
    False otherwise."""
    try:
        return all((i in data for i in expected_value))
    except TypeError:
        return False


def contain_any(data: Iterable[Any], expected_value: Iterable[Any]) -> bool:
    """Returns True if any element in expected_value is present in data,
    False otherwise."""
    try:
        return any((i in data for i in expected_value))
    except TypeError:
        return False


def have_len_equal(data: Sized, expected_value: int) -> bool:
    """Returns True if the length of data is equal to expected_value,
    False otherwise."""
    try:
        return len(data) == expected_value
    except TypeError:
        return False


def have_len_greater(data: Sized, expected_value: int) -> bool:
    """Returns True if the length of data is greater than expected_value,
    False otherwise."""
    try:
        return len(data) > expected_value
    except TypeError:
        return False


def have_len_lesser(data: Sized, expected_value: int) -> bool:
    """Returns True if the length of data is lesser than expected_value,
    False otherwise."""
    try:
        return len(data) < expected_value
    except TypeError:
        return False


Comparator: TypeAlias = Callable[[Any, Any], bool]
COMPARATORS = Literal[
    "is_equal",
    "is_not_equal",
    "is_greater_than",
    "is_lesser_than",
    "is_greater_than_or_equal",
    "is_lesser_than_or_equal",
    "contain",
    "not_contain",
    "contain_all",
    "contain_any",
    "have_len_equal",
    "have_len_greater",
    "have_len_lesser",
]



from typing import Any

import pytest
import requests
from _pytest.logging import LogCaptureFixture
from pytest_mock import MockerFixture

from bepatient import retry
from bepatient.waiter_src.comparators import Comparator
from bepatient.waiter_src.exceptions import WaiterConditionWasNotMet


class TestRetry:
    def test_default_comparator(self, caplog: LogCaptureFixture):
        @retry(1)
        def simple_function():
            return 1

        assert simple_function() == 1
        logs = [
            (
                "bepatient.retry",
                20,
                "Checking whether the condition has been met. The 1 approach."
                " Expected: 1",
            ),
            ("bepatient.retry", 20, "Condition met! Result: 1"),
        ]
        assert caplog.record_tuples == logs

    def test_raise_error(self, caplog: LogCaptureFixture):
        @retry("TEST", loops=1, delay=0)
        def simple_function():
            return 1

        with pytest.raises(WaiterConditionWasNotMet):
            simple_function()
        logs = [
            (
                "bepatient.retry",
                20,
                "Checking whether the condition has been met. The 1 approach."
                " Expected: TEST",
            ),
            ("bepatient.retry", 20, "Condition was not met! Expected: TEST | Result 1"),
        ]
        assert caplog.record_tuples == logs

    @pytest.mark.parametrize(
        "comparator,expected",
        [(lambda a, b: a == b, 200), (lambda a, b: b in a, [201, 200])],
    )
    def test_wait_for_expected(
        self, comparator: Comparator, expected: Any, mocker: MockerFixture
    ):
        @retry(expected=expected, comparer=comparator, delay=0)
        def simple_function() -> int | None:
            try:
                return requests.get("https://webludus.pl", timeout=1).status_code
            except AssertionError:
                return None

        res1 = requests.Response()
        res1.status_code = 404
        res2 = requests.Response()
        res2.status_code = 200

        mocker.patch("requests.get", side_effect=[AssertionError(), res1, res2])
        assert simple_function() == 200
